/*
  This file is under Public Domain.
*/
/*
  Debug Helper
  km.pluginmsg("liveipbar", "UpdateByXPath", "http://kmeleon.sourceforge.net/forum/feed.php?1,type=rss", "/rss/channel/item/title")
 */

#ifdef __MINGW32__
#  define _WIN32_IE 0x0500
#endif

#include <windows.h>

#include <commctrl.h>
#include <iphlpapi.h>

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <tchar.h>
#include <ctype.h>
#include <wctype.h>
#include <wchar.h>

#define KMELEON_PLUGIN_EXPORTS
#include <kmeleon_plugin.h>
#include <KMeleonConst.h>

#include <nscore.h>
#include <nspr.h>
#include <nsNetCID.h>
#include <nsStringAPI.h>
#include <nsEmbedString.h>
#include <nsIWebBrowser.h>
#include <nsIWebBrowserChrome.h>
#include <nsIWebNavigation.h>
#include <nsIEmbeddingSiteWindow.h>

#include <nsIIOService.h>
#include <nsIChannel.h>
#include <nsIDOMParser.h>
#include <nsIDOMDocument.h>
#include <nsIDOMNodeList.h>
#include <nsIDOMXPathEvaluator.h>
#include <nsIDOMXPathResult.h>
#include <nsIInputStream.h>
#include <nsIStreamListener.h>
#include <nsIRequestObserver.h>

#include <nsIPrefService.h>
#include <nsIPrefBranch.h>

#include "miniCOMPtr.h"
#include "miniMBString.h"
#include "miniWString.h"
#include "miniBuffer.h"

#define PLUGIN_NAME TEXT("LiveIPBar plugin")
#define TOOLBAND_NAME TEXT("LiveIPBar")
#define BARPROPNAME TEXT("LiveIPBarProperty")

#define LABEL_TITLE "K-Meleon LiveIP Toolbar"
#define LABEL_CONNECTING "Connecting..."
#define LABEL_DOWNLOADED "Downloaded:"
#define LABEL_NO_SERVICES "ERROR: No services defined."
#define LABEL_TOO_BIG     "ERROR: Contents too big."
#define LABEL_ERROR       "ERROR"

#define PREFKEY_BASENAME               "kmeleon.plugins.liveipbar"
#define PREFKEY_METHOD                 PREFKEY_BASENAME ".method"
#define PREFKEY_SERVICE_PREFIX         PREFKEY_BASENAME ".service"
#define PREFKEY_SERVICE_SUFFIX_URL     ".url"
#define PREFKEY_SERVICE_SUFFIX_TYPE    ".type"
#define PREFKEY_SERVICE_SUFFIX_EXPR    ".expr"

const int MAX_SERVICES = 99;

const size_t DOWNLOAD_LIMIT = 1024 * 256;

/* also used for buffer size */
const size_t MAX_PREFKEY_LENGTH = 64;
const size_t MAX_IP_TEXT_LENGTH = 256;
/* buffer size */
const PRUint32 DOWNLOADED_LABEL_LENGTH = 64;


long DoMessage(const char *, const char *, const char *, long, long);
void Load();
void Create(HWND);
void Close(HWND);
void DoRebar(HWND);
int DoAccel(char *);

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

void PostUpdateMessage();
BOOL DoUpdate(HWND);
BOOL DoUpdateBy(char *);
BOOL DoUpdateByXPath(char *, char *);
void DoGetList(char **);

HWND get_rebar_hwnd(HWND);
HWND get_ipbar_hwnd(HWND);
int message_box(HWND, LPCSTR, LPCSTR, UINT);

WNDPROC g_KMeleonWndProc;
UINT g_id;

kmeleonPlugin kPlugin = {
	KMEL_PLUGIN_VER,
	PLUGIN_NAME,
	DoMessage
};

long DoMessage(const char *to, const char *from, const char *subject,
               long data1, long data2)
{
	if (to[0] == '*' || stricmp(to, kPlugin.dllname) == 0) {
		if (stricmp(subject, "Load") == 0) {
			Load();
		}
		else if (stricmp(subject, "Create") == 0) {
			Create((HWND) data1);
		}
		else if (stricmp(subject, "Close") == 0) {
			Close((HWND) data1);
		}
		else if (stricmp(subject, "DoRebar") == 0) {
			DoRebar((HWND) data1);
		}
		else if (stricmp(subject, "DoAccel") == 0) {
			*(int *)data2 = DoAccel((char *) data1);
		}
		else if (stricmp(subject, "Update") == 0) {
			PostUpdateMessage();
		}
		else if (stricmp(subject, "UpdateBy") == 0) {
			DoUpdateBy((char *) data1);
		}
		else if (stricmp(subject, "UpdateByXPath") == 0) {
			DoUpdateByXPath((char *) data1, (char *)data2);
		}
		else if (stricmp(subject, "GetList") == 0) {
			DoGetList((char **) data2);
		}
		else return 0;
		return 1;
	}
	return 0;
}


void Load()
{
	g_id = kPlugin.kFuncs->GetCommandIDs(1);
}


void Create(HWND hWndParent)
{
	g_KMeleonWndProc = (WNDPROC) SetWindowLongPtr(hWndParent, GWL_WNDPROC,
	                                              (LONG) WndProc);
}


void Close(HWND hWnd)
{
	HWND hwndRebar = get_rebar_hwnd(hWnd);
	if (hwndRebar) {
		RemoveProp(hwndRebar, BARPROPNAME);
	}
}


void DoRebar(HWND rebarWnd)
{
	HWND hwndChild = CreateWindowEx(
		0,
		WC_COMBOBOXEX, TEXT(""),
		WS_CHILD | CBS_DROPDOWNLIST,
		0, 0, 0, 128, rebarWnd, NULL,
		kPlugin.hDllInstance, NULL);
	if (hwndChild == NULL) return;
	SendMessage(hwndChild, CBEM_SETEXSTYLE, 0, 
		CBES_EX_NOEDITIMAGE | CBES_EX_NOEDITIMAGEINDENT);
	
	HFONT hFont;
	hFont = (HFONT) SendMessage(rebarWnd, WM_GETFONT, 0, 0);
	if (hFont != NULL)
		SendMessage(hwndChild, WM_SETFONT, (WPARAM) hFont, TRUE);
	
	HDC dc = GetDC(hwndChild);
	static const LPCTSTR szIP = TEXT("000.000.000.000");
	static const int min = 3 * 4 + 3;
	SIZE size = {16 * min, 16};
	if (dc) {
		GetTextExtentPoint32(dc, szIP, min, &size);
		ReleaseDC(hwndChild, dc);
	}
	
	kPlugin.kFuncs->RegisterBand(hwndChild, TOOLBAND_NAME, TRUE);
	
	REBARBANDINFO rbi = {0};
	rbi.cbSize = sizeof(REBARBANDINFO);
	rbi.fMask  = RBBIM_TEXT | RBBIM_STYLE | RBBIM_CHILD |
		RBBIM_CHILDSIZE | RBBIM_SIZE | RBBIM_IDEALSIZE;
	rbi.lpText = TEXT("LiveIP");
	rbi.fStyle = RBBS_CHILDEDGE | RBBS_FIXEDBMP | RBBS_VARIABLEHEIGHT;
	rbi.hwndChild  = hwndChild;
	rbi.cxMinChild = size.cx;
	rbi.cyMinChild = size.cy + GetSystemMetrics(SM_CXEDGE);
	rbi.cyChild = rbi.cyMinChild;
	rbi.cyMaxChild = rbi.cyMinChild;
	rbi.cyIntegral = 0;
	rbi.cx = rbi.cxMinChild;
	SendMessage(rebarWnd, RB_INSERTBAND, (WPARAM) -1, (LPARAM) &rbi);
	SetProp(rebarWnd, BARPROPNAME, hwndChild);
}


int DoAccel(char *param)
{
	if (stricmp(param, "Update") == 0) {
		return g_id;
	}
	return 0;
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message,
	WPARAM wParam, LPARAM lParam)
{
	switch (message) {
	case WM_COMMAND:
		if (LOWORD(wParam) == g_id) {
			DoUpdate(hWnd);
		}
		break;
	}
	return CallWindowProc(g_KMeleonWndProc, hWnd, message, wParam, lParam);
}


extern "C" {
	BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
	{
		switch (dwReason) {
		case DLL_PROCESS_DETACH:
			break;
		}
		return TRUE;
	}
	
	KMELEON_PLUGIN kmeleonPlugin *GetKmeleonPlugin() {
		return &kPlugin;
	}
	
	KMELEON_PLUGIN int DrawBitmap(DRAWITEMSTRUCT *dis) {
		return 14; // 14 = icon width
	}
}


int message_box(HWND hwnd, LPCTSTR lpText, LPCTSTR lpTitle, UINT style)
{
	if (IsWindowUnicode(hwnd)) {
		return MessageBoxW(hwnd, miniWString(kPlugin.kFuncs->Translate(lpText)), miniWString(kPlugin.kFuncs->Translate(lpTitle)), style);
	}
	return MessageBoxA(hwnd, kPlugin.kFuncs->Translate(lpText), kPlugin.kFuncs->Translate(lpTitle), style);
}


void trace_box(HWND hwnd, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	char *m = PR_vsmprintf(fmt, ap);
	va_end(ap);
	if (m) {
		message_box(hwnd, m, LABEL_TITLE, MB_OK);
		PR_smprintf_free(m);
	}
}


HWND get_frame_hwnd()
{
	nsresult rv;
	
	miniCOMPtr<nsIWebBrowser> browser;
	if (! kPlugin.kFuncs->GetMozillaWebBrowser(NULL, &browser)) {
		return NULL;
	}
	
	miniCOMPtr<nsIWebBrowserChrome> chrome;
	rv = browser->GetContainerWindow(&chrome);
	NS_ENSURE_SUCCESS(rv, NULL);
	NS_ENSURE_TRUE(chrome, NULL);
	
	miniCOMPtr<nsIEmbeddingSiteWindow> esw;
	rv = chrome.QueryInterface(esw);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(esw, NULL);
	
	HWND hwnd = NULL;
	esw->GetSiteWindow((void **) &hwnd);
	
	return hwnd;
}


HWND get_rebar_hwnd(HWND hwnd)
{
	return FindWindowEx(hwnd, NULL, REBARCLASSNAME, NULL);
}


HWND get_ipbar_hwnd(HWND hwnd)
{
	HWND hwndRebar;
	hwndRebar = get_rebar_hwnd(hwnd);
	if (! hwndRebar) return NULL;
	return (HWND) GetProp(hwndRebar, BARPROPNAME);
}


void PostUpdateMessage()
{
	PostMessage(NULL, WM_COMMAND, g_id, 0);
}


void show_message_in_bar(HWND hwnd, LPCTSTR lpText, BOOL trans)
{
	COMBOBOXEXITEM cbei = {0};
	
	SendMessage(hwnd, CB_RESETCONTENT, 0, 0);
	cbei.mask = CBEIF_TEXT;
	cbei.iItem = -1;
	cbei.pszText = const_cast<LPTSTR>(trans ? kPlugin.kFuncs->Translate(lpText) : lpText);
	SendMessage(hwnd, CBEM_INSERTITEM, 0, (LPARAM) &cbei);
	SendMessage(hwnd, CB_SETCURSEL, 0, 0);
}


BOOL update_by_iphelper(HWND hwnd)
{
	PMIB_IPADDRTABLE pIPAddrTable = NULL;
	DWORD dwSize;
	DWORD dwResult;
	DWORD i;
	
	dwSize = sizeof(MIB_IPADDRTABLE);
	
	for (i = 0; i < 2; ++ i) {
		pIPAddrTable = (MIB_IPADDRTABLE *) malloc(dwSize);
		if (! pIPAddrTable) return FALSE;
		dwResult =  GetIpAddrTable(pIPAddrTable, &dwSize, 0);
		if (dwResult == ERROR_SUCCESS) break;
		free(pIPAddrTable);
		if (dwResult != ERROR_INSUFFICIENT_BUFFER) return FALSE;
	}
	if (i == 2) return FALSE;
	
	SendMessage(hwnd, CB_RESETCONTENT, 0, 0);
	for (i = 0; i < pIPAddrTable->dwNumEntries; ++ i) {
		char szIP[MAX_IP_TEXT_LENGTH];
		DWORD dwAddr = pIPAddrTable->table[i].dwAddr;
		wsprintf(szIP, "%u.%u.%u.%u", dwAddr & 0xff, (dwAddr >> 8) & 0xff,
		         (dwAddr >> 16) & 0xff, (dwAddr >> 24) & 0xff);
		COMBOBOXEXITEM cbei = {0};
		cbei.mask = CBEIF_TEXT;
		cbei.iItem = -1;
		cbei.pszText = szIP;
		SendMessage(hwnd, CBEM_INSERTITEM, 0, (LPARAM) &cbei);
	}
	SendMessage(hwnd, CB_SETCURSEL, 0, 0);
	free(pIPAddrTable);
	return TRUE;
}


typedef nsresult (*update_bar_proc)(HWND, const char *, nsresult, PRUint32,
                                    PRUint8 *);


nsresult
update_bar_simple_ascii(HWND hwnd,
                        const char *expr,
                        nsresult status,
                        PRUint32 resultLength,
                        PRUint8 *result)
{
	if (NS_FAILED(status)) {
		show_message_in_bar(hwnd, LABEL_ERROR, TRUE);
		return NS_ERROR_FAILURE;
	}
	
	/*
	   Print until first newline or until max length.
	   Below code expects that result is NULL terminated.
	 */
	LPTSTR text = reinterpret_cast<LPTSTR>(result);
	LPTSTR p;
	for (p = text; *p; p = CharNext(p)) {
		if (*p == TEXT('\r') || *p == TEXT('\n') ||
		    p - text >= MAX_IP_TEXT_LENGTH) {
			*p = TEXT('\0');
			break;
		}
	}
	SendMessage(hwnd, CB_RESETCONTENT, 0, 0);
	COMBOBOXEXITEM cbei = {0};
	cbei.mask = CBEIF_TEXT;
	cbei.iItem = -1;
	cbei.pszText = text;
	SendMessage(hwnd, CBEM_INSERTITEM, 0, (LPARAM) &cbei);
	SendMessage(hwnd, CB_SETCURSEL, 0, 0);
	return NS_OK;
}


void trim_string(nsEmbedString& s)
{
	int n = s.Length() - MAX_IP_TEXT_LENGTH;
	if (n > 0) {
		s.Cut(MAX_IP_TEXT_LENGTH, n);
	}
}


void append_nsstring(HWND hwnd, nsEmbedString& str)
{
	trim_string(str);
	miniMBString text(str.get());
	COMBOBOXEXITEM cbei = {0};
	cbei.mask = CBEIF_TEXT;
	cbei.iItem = -1;
	cbei.pszText = text;
	SendMessage(hwnd, CBEM_INSERTITEM, 0, (LPARAM) &cbei);
}


nsresult append_dom_node(HWND hwnd, miniCOMPtr<nsIDOMNode>& node)
{
	nsresult rv;
	
	PRUint16 nodeType;
	rv = node->GetNodeType(&nodeType);
	NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
	
	if (nodeType == nsIDOMNode::TEXT_NODE ||
	    nodeType == nsIDOMNode::CDATA_SECTION_NODE) {
		nsEmbedString svalue;
		rv = node->GetNodeValue(svalue);
		NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
		
		append_nsstring(hwnd, svalue);
		return NS_OK;
	}
	
	PRBool hasChild;
	rv = node->HasChildNodes(&hasChild);
	NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
	
	if (hasChild) {
		miniCOMPtr<nsIDOMNodeList> nodeList;
		rv = node->GetChildNodes(&nodeList);
		NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
		
		PRUint32 length;
		rv = nodeList->GetLength(&length);
		NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
		
		PRUint32 i;
		for (i = 0; i < length; ++ i) {
			miniCOMPtr<nsIDOMNode> child;
			rv = nodeList->Item(i, &child);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			
			rv = child->GetNodeType(&nodeType);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			
			if (nodeType == nsIDOMNode::TEXT_NODE ||
			    nodeType == nsIDOMNode::CDATA_SECTION_NODE) {
				nsEmbedString svalue;
				rv = child->GetNodeValue(svalue);
				NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
				
				append_nsstring(hwnd, svalue);
			}
		}
	}
	return NS_OK;
}


nsresult
update_bar_xpath(HWND hwnd,
                 const char *expr,
                 nsresult status,
                 PRUint32 resultLength,
                 PRUint8 *result)
{
	nsresult rv;
	
	if (NS_FAILED(status)) {
		show_message_in_bar(hwnd, LABEL_ERROR, TRUE);
		return NS_ERROR_FAILURE;
	}
	
	miniCOMPtr<nsIDOMParser> domparser;
	rv = domparser.doGetService(NS_DOMPARSER_CONTRACTID);
	NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
	NS_ENSURE_TRUE(domparser, NS_ERROR_FAILURE);
	
	miniCOMPtr<nsIDOMDocument> document;
	rv = domparser->ParseFromBuffer(result, resultLength, "text/xml", &document);
	NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
	NS_ENSURE_TRUE(document, NS_ERROR_FAILURE);
	
	miniCOMPtr<nsIDOMXPathEvaluator> evaluator;
	rv = document.QueryInterface(evaluator);
	NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
	NS_ENSURE_TRUE(evaluator, NS_ERROR_FAILURE);
	
	nsEmbedString expression;
	expression = miniWString(expr);
	
	miniCOMPtr<nsIDOMXPathResult> res;
	rv = evaluator->Evaluate(expression, static_cast<nsIDOMDocument *>(document), nsnull, nsIDOMXPathResult::ANY_TYPE, nsnull, reinterpret_cast<nsISupports **>(&res));
	NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
	NS_ENSURE_TRUE(res, NS_ERROR_FAILURE);
	
	PRUint16 type;
	rv = res->GetResultType(&type);
	NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
	SendMessage(hwnd, CB_RESETCONTENT, 0, 0);
	switch (type) {
	case nsIDOMXPathResult::NUMBER_TYPE:
		{
			double fvalue;
			rv = res->GetNumberValue(&fvalue);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			PRInt64 lvalue = NS_STATIC_CAST(PRInt64, fvalue);
			char *svalue;
			if (lvalue != fvalue) {
				svalue = PR_smprintf("%f", fvalue);
			} else {
				svalue = PR_smprintf("%lld", lvalue);
			}
			NS_ENSURE_TRUE(svalue, NS_ERROR_FAILURE);
			
			COMBOBOXEXITEM cbei = {0};
			cbei.mask = CBEIF_TEXT;
			cbei.iItem = -1;
			cbei.pszText = svalue;
			SendMessage(hwnd, CBEM_INSERTITEM, 0, (LPARAM) &cbei);
			
			PR_smprintf_free(svalue);
		}
		break;
	case nsIDOMXPathResult::STRING_TYPE:
		{
			nsEmbedString svalue;
			rv = res->GetStringValue(svalue);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			append_nsstring(hwnd, svalue);
		}
		break;
	case nsIDOMXPathResult::BOOLEAN_TYPE:
		{
			PRBool bvalue;
			rv = res->GetBooleanValue(&bvalue);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			
			COMBOBOXEXITEM cbei = {0};
			cbei.mask = CBEIF_TEXT;
			cbei.iItem = -1;
			cbei.pszText = const_cast<LPTSTR>(bvalue ? "true" : "false");
			SendMessage(hwnd, CBEM_INSERTITEM, 0, (LPARAM) &cbei);
		}
		break;
	case nsIDOMXPathResult::UNORDERED_NODE_ITERATOR_TYPE:
	case nsIDOMXPathResult::ORDERED_NODE_ITERATOR_TYPE:
		for (;;) {
			PRBool state;
			rv = res->GetInvalidIteratorState(&state);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			if (state) break;
			miniCOMPtr<nsIDOMNode> node;
			rv = res->IterateNext(&node);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			if (! node) break;
			append_dom_node(hwnd, node);
		}
		break;
	case nsIDOMXPathResult::UNORDERED_NODE_SNAPSHOT_TYPE:
	case nsIDOMXPathResult::ORDERED_NODE_SNAPSHOT_TYPE:
		{
			PRUint32 length;
			rv = res->GetSnapshotLength(&length);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			PRUint32 i;
			for (i = 0; i < length; i ++) {
				miniCOMPtr<nsIDOMNode> node;
				rv = res->SnapshotItem(i, &node);
				NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
				if (! node) break;
				append_dom_node(hwnd, node);
			}
		}
		break;
	case nsIDOMXPathResult::ANY_UNORDERED_NODE_TYPE:
	case nsIDOMXPathResult::FIRST_ORDERED_NODE_TYPE:
		{
			miniCOMPtr<nsIDOMNode> node;
			rv = res->GetSingleNodeValue(&node);
			NS_ENSURE_SUCCESS(rv, NS_ERROR_FAILURE);
			NS_ENSURE_TRUE(node, NS_ERROR_FAILURE);
			append_dom_node(hwnd, node);
		}
		break;
	default:
		return NS_ERROR_FAILURE;
	}
	SendMessage(hwnd, CB_SETCURSEL, 0, 0);
	return NS_OK;
}


class CStreamListener : public nsIStreamListener
{
public:
	NS_DECL_ISUPPORTS;
	NS_DECL_NSIREQUESTOBSERVER;
	NS_DECL_NSISTREAMLISTENER;
	
	CStreamListener(HWND hwnd, nsEmbedCString& expr, update_bar_proc proc)
		: m_hwnd(hwnd), m_expr(expr), m_proc(proc), m_message(0)
	{
		m_message = kPlugin.kFuncs->Translate(LABEL_DOWNLOADED);
	}
	virtual ~CStreamListener() {}
	
private:
	CStreamListener() {}
	static
	NS_METHOD WriteSegmentFun(nsIInputStream *, void *, const char *,
							  PRUint32, PRUint32, PRUint32 *);
	miniBuffer<char> m_buffer;
	HWND m_hwnd;
	nsEmbedCString m_expr;
	update_bar_proc m_proc;
	LPCTSTR m_message;
};

NS_IMPL_ISUPPORTS1(CStreamListener, nsIStreamListener);

NS_IMETHODIMP
CStreamListener::OnStartRequest(nsIRequest *aRequest,
                                nsISupports *aContext)
{
	miniCOMPtr<nsIChannel> channel;
	channel.doQueryInterface(aRequest);
	if (channel) {
		PRInt32 contentLength = 0;
		channel->GetContentLength(&contentLength);
		/* +1 means space of NULL suffix to ensure NULL terminated. */
		if (contentLength > 0 && m_buffer.SetCapacity(contentLength + 1)) {
			return NS_ERROR_FAILURE;
		}
	}
	return NS_OK;
}

NS_IMETHODIMP
CStreamListener::OnStopRequest(nsIRequest *aRequest,
                               nsISupports *aContext,
                               nsresult aStatusCode)
{
	if (m_buffer.Append('\0')) return NS_ERROR_FAILURE;
	/*
	  -1 means space of NULL suffix.
	  NULL suffix is useful for string operation,
	  but nsIDOMParser.ParseFromBuffer is confused by it.
	 */
	return (*m_proc)(m_hwnd, m_expr.get(), aStatusCode, m_buffer.Length() - 1,
	                 NS_REINTERPRET_CAST(PRUint8 *, m_buffer.Get()));
}

NS_METHOD
CStreamListener::WriteSegmentFun(nsIInputStream *aInputStream,
                void *aClosure,
                const char *aFromSegment,
                PRUint32 aToOffset,
                PRUint32 aCount,
                PRUint32 *aWriteCount)
{
	CStreamListener *self = (CStreamListener *) aClosure;
	
	if (self->m_buffer.Append(aFromSegment, aCount)) {
		return NS_ERROR_FAILURE;
	}
	*aWriteCount = aCount;
	
	return NS_OK;
}

NS_IMETHODIMP
CStreamListener::OnDataAvailable(nsIRequest *aRequest,
                                 nsISupports *aContext,
                                 nsIInputStream *aInputStream,
                                 PRUint32 aOffset,
                                 PRUint32 aCount)
{
	PRUint32 countRead;
	nsresult rv;
	
	rv = aInputStream->ReadSegments(WriteSegmentFun, this, aCount, &countRead);
	if (m_buffer.Length() > DOWNLOAD_LIMIT) {
		show_message_in_bar(m_hwnd, LABEL_TOO_BIG, TRUE);
		return NS_ERROR_FAILURE;
	}
	else {
		TCHAR message[DOWNLOADED_LABEL_LENGTH];
		PR_snprintf(message, DOWNLOADED_LABEL_LENGTH, "%s %u bytes.",
					m_message, m_buffer.Length());
		show_message_in_bar(m_hwnd, message, FALSE);
	}
	return rv;
}


BOOL select_service(nsEmbedCString& url, PRInt32& type, nsEmbedCString& expr)
{
	nsresult rv;
	
	miniCOMPtr<nsIPrefService> prefservice;
	rv = prefservice.doGetService(NS_PREFSERVICE_CONTRACTID);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(prefservice, FALSE);
	
	miniCOMPtr<nsIPrefBranch> prefbranch;
	prefservice->GetBranch(nsnull, &prefbranch);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(prefbranch, FALSE);
	
	int n;
	PRInt32 t;
	char pn[MAX_PREFKEY_LENGTH];
	
	for (n = 0; n < MAX_SERVICES; ++ n) {
		PR_snprintf(pn, MAX_PREFKEY_LENGTH,
					PREFKEY_SERVICE_PREFIX "%d" PREFKEY_SERVICE_SUFFIX_URL,
					n);
		rv = prefbranch->GetPrefType(pn, &t);
		NS_ENSURE_SUCCESS(rv, FALSE);
		if (t != nsIPrefBranch::PREF_STRING) break;
		
		PR_snprintf(pn, MAX_PREFKEY_LENGTH,
					PREFKEY_SERVICE_PREFIX "%d" PREFKEY_SERVICE_SUFFIX_TYPE,
					n);
		rv = prefbranch->GetPrefType(pn, &t);
		NS_ENSURE_SUCCESS(rv, FALSE);
		if (t != nsIPrefBranch::PREF_INT) break;
		
		PR_snprintf(pn, MAX_PREFKEY_LENGTH,
					PREFKEY_SERVICE_PREFIX "%d" PREFKEY_SERVICE_SUFFIX_EXPR,
					n);
		rv = prefbranch->GetPrefType(pn, &t);
		NS_ENSURE_SUCCESS(rv, FALSE);
		if (t != nsIPrefBranch::PREF_STRING) break;
	}
	
	if (n == 0) return FALSE;
	
	int i;
	/* random select */
	i = (rand() >> 8) % n;
	
	PR_snprintf(pn, MAX_PREFKEY_LENGTH,
				PREFKEY_SERVICE_PREFIX "%d" PREFKEY_SERVICE_SUFFIX_URL,
				i);
	rv = prefbranch->GetCharPref(pn, getter_Copies(url));
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	PR_snprintf(pn, MAX_PREFKEY_LENGTH,
				PREFKEY_SERVICE_PREFIX "%d" PREFKEY_SERVICE_SUFFIX_TYPE,
				i);
	rv = prefbranch->GetIntPref(pn, &type);
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	PR_snprintf(pn, MAX_PREFKEY_LENGTH,
				PREFKEY_SERVICE_PREFIX "%d" PREFKEY_SERVICE_SUFFIX_EXPR,
				i);
	rv = prefbranch->GetCharPref(pn, getter_Copies(expr));
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	return TRUE;
}


BOOL update_by_http_service(HWND hwnd, nsEmbedCString& url,
                            PRInt32 type, nsEmbedCString& expr)
{
	nsresult rv;
	
	miniCOMPtr<nsIIOService> ioservice;
	rv = ioservice.doGetService(NS_IOSERVICE_CONTRACTID);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(ioservice, FALSE);
	
	PRBool offline;
	rv = ioservice->GetOffline(&offline);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_FALSE(offline, FALSE);
	
	miniCOMPtr<nsIChannel> channel;
	rv = ioservice->NewChannel(url, nsnull, nsnull, &channel);
	NS_ENSURE_SUCCESS(rv, FALSE);
	NS_ENSURE_TRUE(channel, FALSE);
	
	update_bar_proc proc;
	
	switch (type) {
	case 1:
		proc = update_bar_xpath;
		break;
	default:
		proc = update_bar_simple_ascii;
		break;
	}
	
	miniCOMPtr<CStreamListener> sl(new CStreamListener(hwnd, expr, proc));
	NS_ENSURE_TRUE(sl, FALSE);
	
	rv = channel->AsyncOpen(sl, nsnull);
	NS_ENSURE_SUCCESS(rv, FALSE);
	
	{
		nsEmbedCString message(kPlugin.kFuncs->Translate(LABEL_CONNECTING));
		message += url;
		show_message_in_bar(hwnd, message.get(), FALSE);
	}
	return TRUE;
}


BOOL update_by_http(HWND hwnd)
{
	int n = 0;
	kPlugin.kFuncs->GetPreference(PREF_INT, "kmeleon.general.offline", &n, &n);
	if (n != 0) return FALSE;
	
	nsEmbedCString url, expr;
	PRInt32 type;
	
	if (! select_service(url, type, expr)) {
		show_message_in_bar(hwnd, LABEL_NO_SERVICES, TRUE);
		return FALSE;
	}
	return update_by_http_service(hwnd, url, type, expr);
}


BOOL update(HWND hwnd, int method)
{
	HWND hwndIPBar = get_ipbar_hwnd(hwnd);
	if (! hwndIPBar) return FALSE;
	
	switch (method) {
	case 0:
		return update_by_http(hwndIPBar);
	case 1:
		return update_by_iphelper(hwndIPBar);
	}
	return FALSE;
}


BOOL DoUpdate(HWND hwnd)
{
	if (! hwnd) return FALSE;
	int n = 1;
	kPlugin.kFuncs->GetPreference(PREF_INT, PREFKEY_METHOD, &n, &n);
	return update(hwnd, n);
}


BOOL DoUpdateBy(char *method)
{
	HWND hwnd = get_frame_hwnd();
	if (! hwnd) return FALSE;
	return update(hwnd, atoi(method));
}


BOOL DoUpdateByXPath(char *aUrl, char *aExpr)
{
	HWND hwnd;
	HWND hwndIPBar;
	
	hwnd = get_frame_hwnd();
	if (! hwnd) return FALSE;
	hwndIPBar = get_ipbar_hwnd(hwnd);
	if (! hwndIPBar) return FALSE;
	
	int n = 0;
	kPlugin.kFuncs->GetPreference(PREF_INT, "kmeleon.general.offline", &n, &n);
	if (n != 0) return FALSE;
	
	nsEmbedCString url, expr;
	url = aUrl;
	expr = aExpr;
	
	return update_by_http_service(hwndIPBar, url, 1, expr);
}


void DoGetList(char **result)
{
	HWND hwnd, hwndIPBar;
	nsEmbedCString list;
	if ((hwnd = get_frame_hwnd()) && (hwndIPBar = get_ipbar_hwnd(hwnd))) {
		static const int BUFFER_LENGTH = 256; /* IP is shorter than 256. */
		CHAR buffer[BUFFER_LENGTH];
		for (int i = 0; ; i ++) {
			/* Accoring to the MSDN, after sending CBEM_GETITEM,
			  pszText member of COMBOBOXEXITEM may be changed to
			  point another. Then this initialize on every loop. */
			COMBOBOXEXITEM cbei = {0};
			cbei.mask = CBEIF_TEXT;
			cbei.iItem = i;
			cbei.pszText = buffer;
			cbei.cchTextMax = BUFFER_LENGTH;
			if (! SendMessage(hwndIPBar, CBEM_GETITEM, 0, (LPARAM) &cbei))
				break;
			/* Append (+=) function doesn't return error state,
			   then this assumes success to all. */
			if (i > 0) list += "\r\n";
			list += cbei.pszText; /* Do not use "buffer". */
		}
	}
	*result = _strdup(list.get());
}
