/*
  This file is under Public Domain.
*/

template <class T>
struct miniBuffer
{
	typedef T type;
	miniBuffer() : m_capacity(0), m_length(0), m_buffer(NULL) {}
	~miniBuffer();
	int SetCapacity(size_t);
	int EnsureCapacity(size_t);
	int FitCapacity() { return SetCapacity(Length()); }
	int Append(T);
	int Append(const T*, size_t);
	size_t Length() const { return m_length; }
	T* Get() { return m_buffer; }
	operator T* () { return m_buffer; }
private:
	size_t m_capacity;  /* maximam length of string */ 
	size_t m_length;   /* current length of string */
	T *m_buffer;
};

template <class T>
miniBuffer<T>::~miniBuffer()
{
	free(m_buffer);
}

template <class T>
int miniBuffer<T>::SetCapacity(size_t capacity)
{
	T* p;
	
	p = (T *) realloc(m_buffer, sizeof(T) * (capacity));
	if (! p) return 1;
	m_buffer = p;
	m_capacity = capacity;
	return 0;
}

template <class T>
int miniBuffer<T>::EnsureCapacity(size_t length)
{
	size_t new_length = m_length + length;
	if (new_length < m_length) return 1;
	if (new_length > m_capacity) {
		size_t new_capacity = m_capacity * 2;
		if (new_capacity < new_length) {
			new_capacity = new_length;
		}
		if (SetCapacity(new_capacity)) return 1;
	}
	return 0;
}

template <class T>
int miniBuffer<T>::Append(T item)
{
	if (EnsureCapacity(1)) return 1;
	m_buffer[m_length] = item;
	m_length ++;
	return 0;
}

template <class T>
int miniBuffer<T>::Append(const T* items, size_t length)
{
	if (EnsureCapacity(length)) return 1;
	memcpy(m_buffer + m_length, items, sizeof(T) * length);
	m_length += length;
	return 0;
}
